"""
Client for the Rune Labs graph API.

This package is not intended to be used directly, and implementation is
subject to change. Use the functionality exposed by the `runeq.session()`
client session instead.

"""
